;; PANCAKE-ROM
;; Copyright 2016 - Michael Kohn
;;
;; P50 is an input/output loading and listening to the IR receiver
;; P51 is output to continuous rotation servo
;; P52 is output to regular servero
;;
;; Timer 2 is prescaled / 16
;; 3686400 MHz / 16 = 230400 Hz = 4.3uS
;;
;; Timer 7 is not prescaled
;; 3686400 MHz = 0.27 uS

.65816

.include "w65c265.inc"

.macro PAUSE
.scope
  ldx #0
  stx interrupt_count
  ldx #200
wait:
  cpx interrupt_count
  bpl wait
.ends
.endm

;  ldx #0
;wait:
;  dex
;  bne wait

.macro DEBUG_LED_ON
  ;; Set A to 8-Bit, Debug LED ON, Set A to 16-bit
  sep #0x20
  lda.b #0x00
  sta PD7
  rep #0x20
.endm

.macro DEBUG_LED_OFF
  ;; Set A to 8-Bit, Debug LED OFF, Set A to 16-bit
  sep #0x20
  lda.b #0xff
  sta PD7
  rep #0x20
.endm

.macro WHILE_PANCAKE
.scope
  ;; Wait while sensor is under pancake
  sep #0x20
while_pancake:
  lda PD5
  and.b #1
  beq while_pancake
  ;; Debug LED on because it should be over chocolate now
  lda.b #0x00
  sta PD7
  rep #0x20
.ends
.endm

.macro WHILE_CHOCOLATE
.scope
  ;; Wait while sensor is under chocolate
  sep #0x20
while_chocolate:
  lda PD5
  and.b #1
  bne while_chocolate
  ;; Debug LED off because it should be over pancake now
  lda.b #0xff
  sta PD7
  rep #0x20
.ends
.endm

.define TOP_SERVO_COUNTER 400
;.define TEST_MODE 1

.org 0xf0
count:
  resw 1
noise:
  resw 1
temp:
  resw 1
short:
  resw 1
start_marker:
  resw 1
interrupt_count:
  resw 1
track_timer:
  resw 1
data:
  resw 1
debug:
  resw 1

.org 0x1000
start:
  ; Disable interrupts
  sei

  ; Set native mode
  clc
  xce

  ; Set A to 8-bit
  sep #0x20

  ; Set X/Y to 16-bit
  rep #0x10

  ; UART control: receiver enable, 8 bit, transmit enable
  lda.b #0x25
  sta ACSR0

  ; Write '*' to UART for testing
  ;lda.b #'*'
  ;sta ARTD0

  ; Timer control
  lda.b #0x18
  sta TCR

  ; Enable timer 2, disable timer 0
  lda TER
  ora.b #0x04
  and.b #0xfe
  sta TER

  ; disable interrupts on port 5
  lda.b #0x00
  sta UIER

  ; UART control: receiver enable, 8 bit, transmit enable
  lda.b #0x25
  sta ACSR0

  ; set port 50 as input, 51/52 as output
  lda.b #6
  sta PDD5

  ; set port 5 value to 0
  lda.b #0
  sta PD5

  ; Set timer top value
  ; (3686400 / 16 / 12) = 19200
  ; 1 / 19200 = 0.000052 seconds - approx 50us per interrupt
  ; 16 * 12 cpu cycles per interrupt
  ldx #12
  stx T2LL

  ; Set timer interrupt vector
  ;ldx #timer2_interrupt
  ;stx IRQT2
  ;stx 0x0121
  lda.b #timer2_interrupt & 0xff
  sta 0x0121
  lda.b #timer2_interrupt >> 8
  sta 0x0122
  lda.b #0
  sta 0x0123

  ; Setup track servo timer
  ldx #400-27
  stx track_timer

  ; Enable interrupt for Timer 2
  lda.b #0x04
  sta TIER

  ; Set A to 16-bit
  rep #0x20

  ; Enable interrupts
  cli

main:
  ;; Track 0
  ldx #400-26
  stx track_timer
  ldx #0
  stx start_marker
  PAUSE
  jsr calibrate
  jsr read_track

  ;; Track 1
  ldx #400-27
  stx track_timer
  ldx #0
  stx start_marker
  PAUSE
  jsr calibrate
  jsr read_track

  ;; Track 2
  ldx #400-29
  stx track_timer
  ldx #0
  stx start_marker
  PAUSE
  jsr calibrate
  jsr read_track

  ;; Track 3
  ldx #400-30
  stx track_timer
  ldx #0
  stx start_marker
  PAUSE
  jsr calibrate
  jsr read_track

  ; Disable interrupt for Timer 2
  sep #0x20
  lda.b #0x00
  sta TIER

main_while_1:
  bra main_while_1

calibrate:
  ;; count = 0
  ldx #0
  stx count
calibrate_next_bit:
  WHILE_PANCAKE
  ;; Reset interrupt counter
  ldx #0
  stx interrupt_count
  WHILE_CHOCOLATE

  ;; if interrupt_count > start_marker then start_marker = interrupt_count
  ldx interrupt_count
  cpx start_marker
  bmi calibrate_not_longest
  stx start_marker
calibrate_not_longest:

;debug_list:
;  stx 0x100
;  inc debug_list+1
;  inc debug_list+1

  ;; count++
  ldx count
  inx
  stx count
  ;; if count > 0x80 then break
  cpx #0x80
  bne calibrate_next_bit
  rts

read_track:
  ;; Set up threshold to ignore
  ldx #2
  stx noise
  ;; Set up threshold of "is it a short or long"
  ;; short = (start_marker / 4)
  ;; temp  = (start_marker / 8)
  ;; short = short + temp
  lda start_marker
  lsr
  lsr
  sta short
  lsr
  sta temp
  clc
  adc short
  sta short
  ;; Lower threshold for start_marker
  ;; start_marker = start_marker - temp
  lda start_marker
  sec
  sbc temp
  sta start_marker
  ;; Wait for a chocolate start marker
read_track_wait_long:
  ;; Wait until sensor sees chocolate
  WHILE_PANCAKE
  ;; Reset interrupt counter
  ldx #0
  stx interrupt_count
  ;; Wait while sensor is over chocolate
  WHILE_CHOCOLATE
  ;; if interrupt_count < start_marker then repeat
  ldx interrupt_count
  cpx start_marker
  bmi read_track_wait_long

  ;; Now read data until chocolate mark is start_marker again
  ldx #0
  stx count
read_track_next_bit:
  ;; Wait until sensor sees chocolate
  WHILE_PANCAKE
  ;; Reset interrupt counter
  ldx #0
  stx interrupt_count
  ;; Wait while sensor is over chocolate
  WHILE_CHOCOLATE
  ;; if interrupt_count > start_marker then exit function
  ldx interrupt_count
  cpx start_marker
  bpl read_track_done
  ;; ignore noise
  cpx noise
  bmi read_track_next_bit
  ;; data = data << 1
  asl data
  ;; if interrupt_count > short then data = data + 1
  cpx short
  bmi read_track_is_zero
  inc data
read_track_is_zero:
  ;; count++
  inc count
  ;; if count & 0x7 == 0 then write to UART data
  lda count
  and #0x7
  bne read_track_next_bit
  ;; Set A to 8-Bit, write data over UART, set A to 16-bit
  sep #0x20
  lda data
  sta ARTD0
  rep #0x20
  jmp read_track_next_bit
read_track_done:
  rts

.align 16
timer2_interrupt:
  pha
  php
  ; Set A to 16-bit, increment interrupt_count, Set A to 8-bit
  ;rep #0x20
  ;inc interrupt_count
  ;sep #0x20
  ; ldy, iny, sty seems to be faster (12 cycles) than the above (14 cycles)
  sep #0x20
  ldy interrupt_count
  iny
  sty interrupt_count

  ; Clear interrupt for Timer 2
  lda.b #0x04
  sta TIFR

  ; Update delay count
  ldy delay
  dey
  sty delay
  bne timer2_interrupt_not_0

  ; Reset delay count
  ldy #TOP_SERVO_COUNTER
  sty delay

  ; Set P51 and P52 on
  lda PD5
  ora.b #0x06
  sta PD5
  plp
  pla
  rti

timer2_interrupt_not_0:
  ; 27 = track 0
  ; 28 = track 1
  ; 29 = track 2
  ; 31 = track 3 (inside)
  ;cpy #400-27
  cpy track_timer
  bne timer2_do_not_update_servo1
  lda PD5
  and.b #0x03
  sta PD5
timer2_do_not_update_servo1:
  ;cpy #400-25
  ;cpy #400-35
  cpy #400-30
  ;cpy #400-27   ; very slow
  bne timer2_interrupt_exit
  lda PD5
  and.b #0x05
  ; Comment out to stop spinning
;.ifndef TEST_MODE
  sta PD5
;.endif
timer2_interrupt_exit:
  plp
  pla
  rti

.align 16
delay:
  dw TOP_SERVO_COUNTER

